/*
 * Copyright (C) 2000-2002 Red Hat, Inc.
 *
 * This is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Library General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#include "../config.h"
#include <libintl.h>
#include <locale.h>
#include <pwd.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#include <glade/glade-xml.h>
#include <gtk/gtk.h>
#include <krb5.h>
#include "gui_utils.h"

#ifndef KRB5_SUCCESS
#define KRB5_SUCCESS 0
#endif

#define _(String) dgettext(PACKAGE, String)

static void clear_entry(GtkWidget * widget, GtkWidget * target);

static gchar *cachefile = NULL;
static krb5_context context;

struct flagmap {
	const char *const buttonname;
	guint flag;
};

static const char *const renewtimes[] = {
	"dayspinbuttonr",
	"hourspinbuttonr",
	"minutespinbuttonr"
};

static const char *const lifetimetimes[] = {
	"dayspinbuttonl",
	"hourspinbuttonl",
	"minutespinbuttonl"
};
static void get_flags(GladeXML * xml, krb5_get_init_creds_opt *options)
{
	size_t i;
	gboolean on;
	GtkWidget *widget;
	krb5_timestamp ts;
	widget = glade_xml_get_widget(xml, "forwardablecheckbutton");
	g_object_get(G_OBJECT(widget), "active", &on, NULL);
	krb5_get_init_creds_opt_set_forwardable(options, on);
	
	widget = glade_xml_get_widget(xml, "proxiablecheckbutton");
	g_object_get(G_OBJECT(widget), "active", &on, NULL);
	krb5_get_init_creds_opt_set_proxiable(options, on);
	
	widget = glade_xml_get_widget(xml, "renewablecheckbutton");
	g_object_get(G_OBJECT(widget), "active", &on, NULL);

	ts = get_times(xml,renewtimes) ; 
	krb5_get_init_creds_opt_set_renew_life(options, ts);

	ts = get_times(xml,lifetimetimes);
	if(ts != (krb5_timestamp)0){
		krb5_get_init_creds_opt_set_tkt_life(options, ts);
	}
	
}

krb5_timestamp get_times(GladeXML *xml,const char *const buttons[3])
{
	krb5_timestamp ts = 0;
	GtkWidget *widget;
	widget = glade_xml_get_widget(xml, buttons[0]);
	ts = ((krb5_timestamp)gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget)))*24*60*60;

	widget = glade_xml_get_widget(xml, buttons[1]);
	ts += ((krb5_timestamp)gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget)))*60*60;

	widget = glade_xml_get_widget(xml, buttons[2]);
	ts += ((krb5_timestamp)gtk_spin_button_get_value(GTK_SPIN_BUTTON(widget)))*60;

	return ts;
}


static void renewable_toggled(GtkWidget * button, GladeXML * xml)
{
	GtkWidget *widget;
	gboolean on;
	g_object_get(G_OBJECT(button), "active", &on, NULL);

	widget = glade_xml_get_widget(xml, "dayspinbuttonr");
	gtk_widget_set_sensitive(widget,on);

	widget = glade_xml_get_widget(xml, "hourspinbuttonr");
	gtk_widget_set_sensitive(widget,on);
	
	widget = glade_xml_get_widget(xml, "minutespinbuttonr");
	gtk_widget_set_sensitive(widget,on);
}

static void get_ticket(GtkWidget * button, GladeXML * xml)
{
	GtkWidget *top, entry, *widget;
	krb5_creds creds;
	krb5_get_init_creds_opt options;
	krb5_ccache ccache;
	krb5_flags flags;
	GTimeVal now;
	krb5_principal me;
	int ret;
	const gchar *name;
	const gchar *password;
	const gchar *realm;

	memset(&creds, 0, sizeof(creds));
	top = glade_xml_get_widget(xml, "mainWindow");
	if (!GTK_WIDGET_TOPLEVEL(top)) {
		top = NULL;
	}
	widget = glade_xml_get_widget(xml, "nameentry");
	name = gtk_entry_get_text(GTK_ENTRY(widget));
	widget = glade_xml_get_widget(xml, "passwordentry");
	password = gtk_entry_get_text(GTK_ENTRY(widget));
	widget = glade_xml_get_widget(xml, "realmcombo-entry");
	realm = gtk_entry_get_text(GTK_ENTRY(widget));
	if ((name == NULL) || (strlen(name) == 0)) {
		const char *noprincmessages[] = {	// Predictable applications are no fun.
			_("Neuro implant failed.\nCan't transfer the principal name from your brain."),
			_("Please specify a principal name"),
			_("No principal name given")
		};
		show_error(top, noprincmessages[g_random_int_range(0, sizeof(noprincmessages) / sizeof(noprincmessages[0]))]);
		return;
	}


	if ((strchr(name, '@') == NULL) && (realm != NULL) && (strlen(realm) > 0)) {
		gchar *tmp = g_strdup_printf("%s@%s", name, realm);
		ret = krb5_parse_name(context, tmp, &me);
		g_free(tmp);
	} else {
		//uhm.. get a default realm ?
		ret = krb5_parse_name(context, name, &me);
	}

	if ((password == NULL) || (strlen(password) == 0)) {
		show_error(top, _("No password entered!"));
		krb5_free_principal(context,me);

		return;
	}
	ret = krb5_cc_resolve(context, cachefile, &ccache);
	if (ret != KRB5_SUCCESS) {
		char *msg = g_strdup_printf(_("Error resolving cache name:\n%s"),error_message(ret));
		show_error(top, msg);
		g_free(msg);
		goto out;
	}
	krb5_get_init_creds_opt_init(&options);
	get_flags(xml, &options);
	g_get_current_time(&now);
	ret = krb5_get_init_creds_password(context, &creds, me, (char *)password, NULL, 0,0, NULL, &options );
	if (ret != KRB5_SUCCESS) {
		char *msg;
		msg = g_strdup_printf(_("Error getting credentials:\n%s"), error_message(ret));
		show_error(top, msg);
		g_free(msg);
	} else {
		GtkWidget *dialog;
		dialog = gtk_message_dialog_new(GTK_WINDOW(top), GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT, GTK_MESSAGE_INFO, GTK_BUTTONS_CLOSE, _("Credentials successfully aquired"));
		gtk_dialog_run(GTK_DIALOG(dialog));
		gtk_widget_destroy(dialog);
		//initialize ccache and store new credentials
		ret = krb5_cc_initialize(context, ccache, me);
		if (ret != KRB5_SUCCESS) {
			char *msg;
			msg = g_strdup_printf(_("Error creating credential cache:\n%s"), error_message(ret));
			show_error(top, msg);
			g_free(msg);
			goto out;
		}
		ret = krb5_cc_store_cred(context, ccache, &creds);
		if (ret != KRB5_SUCCESS) {
			char *msg;
			msg = g_strdup_printf(_("Error storing credentials:\n%s"), error_message(ret));
			show_error(top, msg);
			g_free(msg);
			goto out;
		}


	}

out:
	if(creds.client == me){
		creds.client = 0;
	}
	krb5_free_principal(context,me);
	krb5_free_cred_contents(context, &creds);
	krb5_cc_close(context, ccache);
	widget = glade_xml_get_widget(xml, "passwordentry");
	clear_entry(widget, widget);

}

static GValueArray *get_realms(void)
{
	void *iterator;
	char *arealm;
	GValueArray *ret;
	GValue value;
	ret = g_value_array_new(1);
	memset(&value, 0, sizeof(value));
	if (krb5_realm_iterator_create(context, &iterator) == KRB5_SUCCESS) {
		while (krb5_realm_iterator(context, &iterator, &arealm) == KRB5_SUCCESS) {
			if (arealm != NULL) {
				g_value_init(&value, G_TYPE_STRING);
				g_value_set_string(&value, arealm);
				krb5_free_realm_string(context, arealm);
				g_value_array_append(ret, &value);
				g_value_unset(&value);
			}
		}
		krb5_realm_iterator_free(context, &iterator);
	}
	return ret;
}

static gboolean destroy_mainwindow(GtkWidget * widget, GdkEventAny * event, GladeXML * xml)
{
	gtk_widget_destroy(widget);
	g_object_unref(G_OBJECT(xml));
	krb5_free_context(context);
	gtk_main_quit();
	return TRUE;

}
static void clear_entry(GtkWidget * widget, GtkWidget * target)
{
	g_assert(GTK_IS_ENTRY(target));
	gtk_entry_set_text(GTK_ENTRY(target), "");
}


static void init_and_show_gui()
{
	GladeXML *xml;
	GtkWidget *widget;
	krb5_principal princ;
	krb5_ccache ccache;
	GtkListStore *liststore = NULL;
	GValueArray *realms;
	gchar *realm;
	gchar *p, *principal;
	int ret;
	/* Load the GUI. */
	xml = glade_xml_new(DATADIR "/" PACKAGE "/gkinit.glade", "mainWindow", PACKAGE);
	if (xml == NULL) {
		xml = glade_xml_new("gkinit.glade", "mainWindow", PACKAGE);
	}
	if (xml == NULL) {
		show_error(NULL, _("Error initializing GUI!"));
		exit(1);
	}
	ret = krb5_cc_resolve(context, cachefile, &ccache);
	if (ret != KRB5_SUCCESS) {
		gchar *err = g_strdup_printf(_("Error resolving cache name:\n%s"), error_message(ret));
		show_error(glade_xml_get_widget(xml, "mainWindow"), err);
		g_free(err);
		exit(1);
	}
	/* Get the default principal's name . */
	ret = krb5_cc_get_principal(context, ccache, &princ);
	if (ret == KRB5_SUCCESS) {
		gchar *tmp;
		ret = krb5_unparse_name(context, princ, &tmp);
		if (ret == KRB5_SUCCESS) {
			principal = g_strdup(tmp);
			krb5_free_unparsed_name(context, tmp);
			krb5_free_principal(context, princ);
		} else {
			principal = g_strdup("");
		}
	} else {
		struct passwd *pwd;
		setpwent();
		pwd = getpwuid(getuid());
		if (pwd != NULL) {
			principal = g_strdup(pwd->pw_name);
		} else {	//system is b0rked
			principal = g_strdup("");
		}
		endpwent();
	}
	//ahow only "user" name
	if ((p = strstr(principal, "@")) != NULL) {
		*p = '\0';
		p++;
		realm = p;
	} else {
		realm = NULL;
	}
	widget = glade_xml_get_widget(xml, "nameentry");
	if (GTK_IS_ENTRY(widget)) {
		gtk_entry_set_text(GTK_ENTRY(widget), principal);
	}
	//initialize realms combo
	realms = get_realms();
	widget = glade_xml_get_widget(xml, "realmcombo");
	if (realms->n_values > 0) {
		guint i;
		GValue *value;
		GList *list = NULL;
		for (i = 0; i < realms->n_values; i++) {
			value = g_value_array_get_nth(realms, i);
			list = g_list_append(list, (void *) g_value_get_string(value));
		}
		gtk_combo_set_popdown_strings(GTK_COMBO(widget), list);
		if (realm != NULL) {
			widget = (GTK_COMBO(widget))->entry;
			if (GTK_IS_ENTRY(widget)) {
				gtk_entry_set_text(GTK_ENTRY(widget), realm);
			}
		}
		if (list != NULL) {
			g_list_free(list);
		}
		g_value_array_free(realms);
	}
	g_free(principal);

	krb5_cc_close(context, ccache);

	widget = glade_xml_get_widget(xml, "getbutton");
	g_signal_connect(G_OBJECT(widget), "clicked", G_CALLBACK(get_ticket), xml);
	widget = glade_xml_get_widget(xml, "renewablecheckbutton");
	g_signal_connect(G_OBJECT(widget), "clicked", G_CALLBACK(renewable_toggled), xml);
	/* Add an extra callback to clear the password-entry field. */
	widget = glade_xml_get_widget(xml, "passwordentry");
	g_signal_connect(G_OBJECT(widget), "grab-focus", G_CALLBACK(clear_entry), widget);
	/* Show the main window. */
	widget = glade_xml_get_widget(xml, "mainWindow");
	g_signal_connect(G_OBJECT(widget), "delete_event", G_CALLBACK(destroy_mainwindow), xml);
	gtk_widget_show_all(widget);

}

int main(int argc, char **argv)
{
	int ret;

	/* Initialize the locale settings. */
	setlocale(LC_ALL, "");
	bindtextdomain(PACKAGE, GNOMELOCALEDIR);
	textdomain(PACKAGE);

	/* Initialize GTK+. */
	g_random_set_seed((guint32) getpid());
	gtk_init(&argc, &argv);
	/* Initialize Kerberos. */
	initialize_krb5_error_table();
	if ((ret = krb5_init_context(&context)) != KRB5_SUCCESS) {
		gchar *p;
		p = g_strdup_printf(_("Error initializing Kerberos:\n%s"), error_message(ret));
		show_error(NULL, p);
		g_free(p);
		return 1;
	}

	cachefile = g_strdup(krb5_cc_default_name(context));
	ret = krb5_cc_set_default_name(context, cachefile);
	if (ret != KRB5_SUCCESS) {
		gchar *p;
		p = g_strdup_printf(_("Error setting default cache name:\n%s"), error_message(ret));
		show_error(NULL, p);
		g_free(p);
		return 1;
	}

	init_and_show_gui();

	gtk_main();

	return 0;
}

